from __future__ import annotations

import json
from http import HTTPStatus

import httpx

from .parser import FalClientParser, get_output_parser


def _queue_size(args):
    from fal.api.client import SyncServerlessClient
    from fal.api.deploy import _get_user

    client = SyncServerlessClient(host=args.host, team=args.team)._create_rest_client()
    user = _get_user(client)

    url = f"{client.base_url}/applications/{user.username}/{args.app_name}/queue"
    headers = client.get_headers()

    with httpx.Client(base_url=client.base_url, headers=headers, timeout=300) as c:
        resp = c.get(url)

    if resp.status_code != HTTPStatus.OK:
        try:
            detail = resp.json().get("detail", resp.text)
        except Exception:
            detail = resp.text
        raise RuntimeError(f"Failed to get queue size: {detail}")

    data = resp.json()
    size = data.get("size", 0)

    if args.output == "json":
        args.console.print(json.dumps({"size": size}))
    else:
        args.console.print(f"Queue size: {size}")


def _queue_flush(args):
    from fal.api.client import SyncServerlessClient
    from fal.api.deploy import _get_user

    client = SyncServerlessClient(host=args.host, team=args.team)._create_rest_client()
    user = _get_user(client)

    url = f"{client.base_url}/applications/{user.username}/{args.app_name}/queue"
    headers = client.get_headers()

    with httpx.Client(base_url=client.base_url, headers=headers, timeout=300) as c:
        resp = c.delete(url)

    if resp.status_code != HTTPStatus.OK:
        try:
            detail = resp.json().get("detail", resp.text)
        except Exception:
            detail = resp.text
        raise RuntimeError(f"Failed to flush queue: {detail}")

    args.console.print("Queue flushed successfully.")


def add_parser(main_subparsers, parents):
    queue_help = "Manage application queues."
    parser = main_subparsers.add_parser(
        "queue",
        description=queue_help,
        help=queue_help,
        parents=parents,
    )

    subparsers = parser.add_subparsers(
        title="Commands",
        metavar="command",
        required=True,
        parser_class=FalClientParser,
    )

    size_help = "Get queue size for an application."
    size_parser = subparsers.add_parser(
        "size",
        description=size_help,
        help=size_help,
        parents=[*parents, get_output_parser()],
    )
    size_parser.add_argument(
        "app_name",
        help="Application name (do not prefix with owner).",
    )
    size_parser.set_defaults(func=_queue_size)

    flush_help = "Flush all pending requests in an application queue."
    flush_parser = subparsers.add_parser(
        "flush",
        description=flush_help,
        help=flush_help,
        parents=parents,
    )
    flush_parser.add_argument(
        "app_name",
        help="Application name.",
    )
    flush_parser.set_defaults(func=_queue_flush)



from contextlib import asynccontextmanager

from anyio import create_task_group
from fastapi import Request


@asynccontextmanager
async def cancel_on_disconnect(request: Request):
    """
    Async context manager for async code that needs to be cancelled if client
    disconnects prematurely.
    The client disconnect is monitored through the Request object.
    """
    async with create_task_group() as tg:

        async def watch_disconnect():
            while True:
                message = await request.receive()

                if message["type"] == "http.disconnect":
                    tg.cancel_scope.cancel()
                    break

        tg.start_soon(watch_disconnect)

        try:
            yield
        finally:
            tg.cancel_scope.cancel()



from __future__ import annotations

import os

from fal.config import Config


def bool_envvar(name: str):
    if name in os.environ:
        val = os.environ[name].lower().strip()
        return val != "0" and val != "false" and val != ""
    return False


DEBUG = bool_envvar("DEBUG")
TEST_MODE = bool_envvar("ISOLATE_TEST_MODE")
AUTH_DISABLED = bool_envvar("ISOLATE_AUTH_DISABLED")

config = Config()
config_host = config.get("host")

# Explicit host overrides (highest precedence)
_GRPC_HOST_ENV = os.getenv("FAL_GRPC_HOST")
_REST_HOST_ENV = os.getenv("FAL_REST_HOST")
_RUN_HOST_ENV = os.getenv("FAL_RUN_HOST")
_QUEUE_RUN_HOST_ENV = os.getenv("FAL_QUEUE_RUN_HOST")

# Legacy: FAL_HOST takes precedence over config.host (unless FAL_GRPC_HOST is set)
GRPC_HOST = _GRPC_HOST_ENV or os.getenv("FAL_HOST") or config_host or "api.alpha.fal.ai"
if not TEST_MODE:
    # Keep legacy safety check, but allow explicit overrides via FAL_GRPC_HOST.
    if not _GRPC_HOST_ENV:
        assert GRPC_HOST.startswith(
            "api"
        ), "FAL_HOST must start with 'api' (use FAL_GRPC_HOST to override)"

REST_HOST = _REST_HOST_ENV or (
    GRPC_HOST.replace("api", "rest", 1) if GRPC_HOST.startswith("api") else GRPC_HOST
)
REST_SCHEME = "http" if TEST_MODE or AUTH_DISABLED else "https"
REST_URL = f"{REST_SCHEME}://{REST_HOST}"

# fal.run / env.fal.run
FAL_RUN_HOST = _RUN_HOST_ENV or (
    GRPC_HOST.replace("api.", "", 1).replace("alpha.", "", 1).replace(".ai", ".run", 1)
)

FAL_QUEUE_RUN_HOST = _QUEUE_RUN_HOST_ENV or f"queue.{FAL_RUN_HOST}"

DONT_OPEN_LINKS = bool_envvar("FAL_DONT_OPEN_LINKS")

# Controls which URLs are printed when running/deploying apps
# Valid values: "all" (default), "playground", "none"
URL_OUTPUT = os.getenv("FAL_URL_OUTPUT", "all").lower().strip()



from __future__ import annotations

import os

import pytest

import fal
from fal import App, endpoint
from fal.container import ContainerImage


def test_app_regions_propagate_to_function_options():
    from fal.app import wrap_app

    class RegionsApp(App):
        regions = ["us-east", "eu-west"]

        @endpoint("/")
        def hello(self) -> str:
            return "Hello, world!"

    fn = wrap_app(RegionsApp)
    assert fn.options.host.get("regions") == ["us-east", "eu-west"]


def test_app_default_app_name_is_generated_from_class_name():
    class MyCustomApp(App):
        pass

    assert MyCustomApp.app_name == "my-custom-app"


def test_app_classvars_propagate_to_host_kwargs():
    class VarsApp(App):
        request_timeout = 11
        startup_timeout = 22
        min_concurrency = 2
        max_concurrency = 3
        concurrency_buffer = 4
        concurrency_buffer_perc = 50
        scaling_delay = 33
        max_multiplexing = 7
        kind = "container"
        image = ContainerImage.from_dockerfile_str(
            "FROM python:3.10-slim",
        )

    hk = VarsApp.host_kwargs
    assert hk["request_timeout"] == 11
    assert hk["startup_timeout"] == 22
    assert hk["min_concurrency"] == 2
    assert hk["max_concurrency"] == 3
    assert hk["concurrency_buffer"] == 4
    assert hk["concurrency_buffer_perc"] == 50
    assert hk["scaling_delay"] == 33
    assert hk["max_multiplexing"] == 7
    assert hk["kind"] == "container"
    assert isinstance(hk["image"], ContainerImage)


def test_app_files_classvars_propagate_to_host_kwargs():
    class VarsApp(App):
        request_timeout = 11
        startup_timeout = 22
        app_files = ["a.py", "b.py"]
        app_files_ignore = [r"\\.venv/"]
        app_files_context_dir = "."
        min_concurrency = 2
        max_concurrency = 3
        concurrency_buffer = 4
        concurrency_buffer_perc = 50
        scaling_delay = 33
        max_multiplexing = 7

    hk = VarsApp.host_kwargs
    assert hk["request_timeout"] == 11
    assert hk["startup_timeout"] == 22
    assert hk["app_files"] == ["a.py", "b.py"]
    assert hk["app_files_ignore"] == [r"\\.venv/"]
    assert hk["app_files_context_dir"] == "."
    assert hk["min_concurrency"] == 2
    assert hk["max_concurrency"] == 3
    assert hk["concurrency_buffer"] == 4
    assert hk["concurrency_buffer_perc"] == 50
    assert hk["scaling_delay"] == 33
    assert hk["max_multiplexing"] == 7


def test_app_kwargs_merge_into_host_kwargs_and_override_defaults():
    class KwargsApp(App, keep_alive=123, resolver="pip", _scheduler="nomad"):
        pass

    hk = KwargsApp.host_kwargs
    assert hk["keep_alive"] == 123
    assert hk["resolver"] == "pip"
    assert hk["_scheduler"] == "nomad"
    # default keys should still exist
    assert "_scheduler_options" in hk


def test_non_host_classvars_do_not_leak_into_host_kwargs():
    class LeakCheckApp(App):
        requirements = ["fastapi"]
        local_python_modules = ["."]
        machine_type = "M"
        num_gpus = 1
        app_auth = "private"

    hk = LeakCheckApp.host_kwargs
    assert "requirements" not in hk
    assert "local_python_modules" not in hk
    assert "machine_type" not in hk
    assert "num_gpus" not in hk
    assert "app_auth" not in hk


@pytest.mark.asyncio
async def test_runner_state_lifecycle_complete():
    """Test that FAL_RUNNER_STATE transitions through all phases correctly"""
    states = []

    class StateCheckApp(App):
        def setup(self):
            states.append(("setup", os.getenv("FAL_RUNNER_STATE")))

        def teardown(self):
            states.append(("teardown", os.getenv("FAL_RUNNER_STATE")))

    app = StateCheckApp(_allow_init=True)

    # Create a mock FastAPI app
    import fastapi

    fastapi_app = fastapi.FastAPI()

    async with app.lifespan(fastapi_app):
        states.append(("running", os.getenv("FAL_RUNNER_STATE")))

    # Verify the full lifecycle
    assert len(states) == 3
    assert states[0] == ("setup", "SETUP")
    assert states[1] == ("running", "RUNNING")
    assert states[2] == ("teardown", "STOPPING")


def test_function_decorator_rejects_app_files_with_container_kind():
    """Test that app_files cannot be used with kind='container'."""
    image = ContainerImage.from_dockerfile_str("FROM python:3.11-slim")

    error_message = "app_files is not supported for container apps"
    with pytest.raises(ValueError, match=error_message):

        @fal.function("container", app_files=["a.py"])
        def container_because_kind_is_container():
            pass

    with pytest.raises(ValueError, match=error_message):

        @fal.function(image=image, app_files=["a.py"])
        def container_because_image_is_provided():
            pass



from dataclasses import dataclass, field
from typing import Dict, Literal, Optional

Builder = Literal["depot", "service", "worker"]
BUILDERS = {"depot", "service", "worker"}
DEFAULT_COMPRESSION: str = "gzip"
DEFAULT_FORCE_COMPRESSION: bool = False


@dataclass
class ContainerImage:
    """ContainerImage represents a Docker image that can be built
    from a Dockerfile.
    """

    dockerfile_str: str
    build_args: Dict[str, str] = field(default_factory=dict)
    registries: Dict[str, Dict[str, str]] = field(default_factory=dict)
    builder: Optional[Builder] = field(default=None)
    compression: str = DEFAULT_COMPRESSION
    force_compression: bool = DEFAULT_FORCE_COMPRESSION
    secrets: Dict[str, str] = field(default_factory=dict)

    def __post_init__(self) -> None:
        if self.registries:
            for registry in self.registries.values():
                keys = registry.keys()
                if "username" not in keys or "password" not in keys:
                    raise ValueError(
                        "Username and password are required for each registry"
                    )

        if self.builder and self.builder not in BUILDERS:
            raise ValueError(
                f"Invalid builder: {self.builder}, must be one of {BUILDERS}"
            )

    @classmethod
    def from_dockerfile_str(cls, text: str, **kwargs) -> "ContainerImage":
        return cls(dockerfile_str=text, **kwargs)

    @classmethod
    def from_dockerfile(cls, path: str, **kwargs) -> "ContainerImage":
        with open(path) as fobj:
            return cls.from_dockerfile_str(fobj.read(), **kwargs)

    def to_dict(self) -> dict:
        return {
            "dockerfile_str": self.dockerfile_str,
            "build_args": self.build_args,
            "registries": self.registries,
            "builder": self.builder,
            "compression": self.compression,
            "force_compression": self.force_compression,
            "secrets": self.secrets,
        }



from __future__ import annotations

import json
import math
import os
import threading
from base64 import b64encode
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, Generator, Generic, TypeVar
from urllib.error import HTTPError, URLError
from urllib.parse import urlparse, urlunparse
from urllib.request import Request, urlopen
from urllib.response import addinfourl

from fal.auth import key_credentials
from fal.toolkit.exceptions import FileUploadException
from fal.toolkit.file.types import FileData, FileRepository
from fal.toolkit.utils.retry import retry

_FAL_CDN = "https://fal.media"
_FAL_CDN_V3 = "https://v3.fal.media"

DEFAULT_REQUEST_TIMEOUT = 10
PUT_REQUEST_TIMEOUT = 5 * 60

MAX_ATTEMPTS = 5
BASE_DELAY = 0.1
MAX_DELAY = 30
RETRY_CODES = [408, 409, 429, 500, 502, 503, 504]


@contextmanager
def _urlopen(
    request: Request,
    timeout: int = DEFAULT_REQUEST_TIMEOUT,
) -> Generator[addinfourl, None, None]:
    with urlopen(request, timeout=timeout) as response:
        yield response


def _should_retry(exc: Exception) -> bool:
    if isinstance(exc, HTTPError) and exc.code in RETRY_CODES:
        return True
    elif type(exc) is URLError:
        # URLError is a base class for other errors,
        # but it can be raised directly, e.g.
        # URLError: <urlopen error [Errno -3] Temporary failure in name resolution>
        return True

    if isinstance(exc, TimeoutError):
        return True

    return False


@contextmanager
def _maybe_retry_request(
    request: Request,
    **kwargs: Any,
) -> Generator[addinfourl, None, None]:
    _urlopen_with_retry = retry(
        max_retries=MAX_ATTEMPTS,
        base_delay=BASE_DELAY,
        max_delay=MAX_DELAY,
        backoff_type="exponential",
        jitter=True,
        should_retry=_should_retry,
    )(_urlopen)

    with _urlopen_with_retry(request, **kwargs) as response:
        yield response


def _object_lifecycle_headers(
    headers: dict[str, str],
    object_lifecycle_preference: dict[str, str] | None,
):
    if object_lifecycle_preference:
        # Used by V3 CDN
        headers["X-Fal-Object-Lifecycle"] = json.dumps(object_lifecycle_preference)
        # Used by V1 CDN
        headers["X-Fal-Object-Lifecycle-Preference"] = json.dumps(
            object_lifecycle_preference
        )


@dataclass
class FalV2Token:
    token: str
    token_type: str
    base_upload_url: str
    expires_at: datetime

    def is_expired(self) -> bool:
        return datetime.now(timezone.utc) >= self.expires_at


class FalV3Token(FalV2Token):
    pass


class FalV2TokenManager:
    token_cls: type[FalV2Token] = FalV2Token
    storage_type: str = "fal-cdn"
    upload_prefix = "upload."

    def __init__(self):
        self._token: FalV2Token = self.token_cls(
            token="",
            token_type="",
            base_upload_url="",
            expires_at=datetime.min.replace(tzinfo=timezone.utc),
        )
        self._lock: threading.Lock = threading.Lock()

    def get_token(self) -> FalV2Token:
        with self._lock:
            if self._token.is_expired():
                self._refresh_token()
            return self._token

    def _refresh_token(self) -> None:
        key_creds = key_credentials()
        if not key_creds:
            raise FileUploadException("FAL_KEY must be set")

        key_id, key_secret = key_creds
        headers = {
            "Authorization": f"Key {key_id}:{key_secret}",
            "Accept": "application/json",
            "Content-Type": "application/json",
        }

        grpc_host = os.environ.get("FAL_HOST", "api.alpha.fal.ai")
        rest_host = grpc_host.replace("api", "rest", 1)
        url = f"https://{rest_host}/storage/auth/token?storage_type={self.storage_type}"

        req = Request(
            url,
            headers=headers,
            data=b"{}",
            method="POST",
        )
        with _maybe_retry_request(req) as response:
            result = json.load(response)

        parsed_base_url = urlparse(result["base_url"])
        base_upload_url = urlunparse(
            parsed_base_url._replace(netloc=self.upload_prefix + parsed_base_url.netloc)
        )

        self._token = self.token_cls(
            token=result["token"],
            token_type=result["token_type"],
            base_upload_url=base_upload_url,
            expires_at=datetime.fromisoformat(result["expires_at"]),
        )

    def __getstate__(self) -> dict[str, Any]:
        state = self.__dict__.copy()
        # Remove the lock from the state dictionary
        del state["_lock"]
        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        # Restore the instance attributes
        self.__dict__.update(state)
        # Recreate the lock
        self._lock = threading.Lock()


class FalV3TokenManager(FalV2TokenManager):
    token_cls: type[FalV2Token] = FalV3Token
    storage_type: str = "fal-cdn-v3"
    upload_prefix = ""


fal_v2_token_manager = FalV2TokenManager()
fal_v3_token_manager = FalV3TokenManager()


VariableType = TypeVar("VariableType")


class VariableReference(Generic[VariableType]):
    def __init__(self, value: VariableType) -> None:
        self.set(value)

    def get(self) -> VariableType:
        return self.value

    def set(self, value: VariableType) -> None:
        self.value = value


LIFECYCLE_PREFERENCE: VariableReference[dict[str, str] | None] = VariableReference(None)


@dataclass
class FalFileRepositoryBase(FileRepository):
    def _save(
        self, file: FileData, storage_type: str, headers: dict[str, str] | None = None
    ) -> str:
        key_creds = key_credentials()
        if not key_creds:
            raise FileUploadException("FAL_KEY must be set")

        key_id, key_secret = key_creds
        headers = {
            "Authorization": f"Key {key_id}:{key_secret}",
            "Accept": "application/json",
            "Content-Type": "application/json",
            **(headers or {}),
        }

        grpc_host = os.environ.get("FAL_HOST", "api.alpha.fal.ai")
        rest_host = grpc_host.replace("api", "rest", 1)
        storage_url = (
            f"https://{rest_host}/storage/upload/initiate?storage_type={storage_type}"
        )

        try:
            req = Request(
                storage_url,
                data=json.dumps(
                    {
                        "file_name": file.file_name,
                        "content_type": file.content_type,
                    }
                ).encode(),
                headers=headers,
                method="POST",
            )
            with _maybe_retry_request(req) as response:
                result = json.load(response)

            upload_url = result["upload_url"]
        except HTTPError as e:
            raise FileUploadException(
                f"Error initiating upload. Status {e.status}: {e.reason}"
            )

        try:
            req = Request(
                upload_url,
                method="PUT",
                data=file.data,
                headers={"Content-Type": file.content_type},
            )

            with _maybe_retry_request(req, timeout=PUT_REQUEST_TIMEOUT):
                pass

            return result["file_url"]
        except HTTPError as e:
            raise FileUploadException(
                f"Error uploading file. Status {e.status}: {e.reason}"
            )


class MultipartUploadGCS:
    MULTIPART_THRESHOLD = 100 * 1024 * 1024
    MULTIPART_CHUNK_SIZE = 10 * 1024 * 1024
    MULTIPART_MAX_CONCURRENCY = 10

    def __init__(
        self,
        file_name: str,
        chunk_size: int | None = None,
        content_type: str | None = None,
        max_concurrency: int | None = None,
    ) -> None:
        self.file_name = file_name
        self.chunk_size = chunk_size or self.MULTIPART_CHUNK_SIZE
        self.content_type = content_type or "application/octet-stream"
        self.max_concurrency = max_concurrency or self.MULTIPART_MAX_CONCURRENCY

        self._access_url: str | None = None
        self._upload_url: str | None = None

        self._parts: list[dict] = []

    @property
    def access_url(self) -> str:
        if not self._access_url:
            raise FileUploadException("Upload not initiated")
        return self._access_url

    @property
    def upload_url(self) -> str:
        if not self._upload_url:
            raise FileUploadException("Upload not initiated")
        return self._upload_url

    @property
    def auth_headers(self) -> dict[str, str]:
        fal_key = key_credentials()
        if not fal_key:
            raise FileUploadException("FAL_KEY must be set")

        key_id, key_secret = fal_key
        return {
            "Authorization": f"Key {key_id}:{key_secret}",
        }

    def create(self):
        grpc_host = os.environ.get("FAL_HOST", "api.alpha.fal.ai")
        rest_host = grpc_host.replace("api", "rest", 1)
        url = f"https://{rest_host}/storage/upload/initiate-multipart?storage_type=gcs"

        try:
            req = Request(
                url,
                method="POST",
                headers={
                    **self.auth_headers,
                    "Content-Type": "application/json",
                    "Accept": "application/json",
                },
                data=json.dumps(
                    {
                        "file_name": self.file_name,
                        "content_type": self.content_type,
                    }
                ).encode(),
            )

            with _maybe_retry_request(req) as response:
                result = json.load(response)
                self._access_url = result["file_url"]
                self._upload_url = result["upload_url"]

        except HTTPError as exc:
            raise FileUploadException(
                f"Error initiating upload. Status {exc.status}: {exc.reason}"
            )

    def upload_part(self, part_number: int, data: bytes) -> None:
        initiate_upload_url = self.upload_url + f"/{part_number}"
        req = Request(
            initiate_upload_url,
            method="POST",
            headers=self.auth_headers,
        )

        try:
            with _maybe_retry_request(req) as response:
                result = json.load(response)
                upload_url = result["upload_url"]
        except HTTPError as exc:
            raise FileUploadException(
                f"Error initiating part {part_number} to {initiate_upload_url}. "
                f"Status {exc.status}: {exc.reason}"
            )

        req = Request(
            upload_url,
            method="PUT",
            data=data,
        )

        try:
            with _maybe_retry_request(req, timeout=PUT_REQUEST_TIMEOUT) as resp:
                self._parts.append(
                    {
                        "part_number": part_number,
                        "etag": resp.headers["ETag"].strip('"'),
                    }
                )
        except HTTPError as exc:
            raise FileUploadException(
                f"Error uploading part {part_number} to {upload_url}. "
                f"Status {exc.status}: {exc.reason}"
            )

    def complete(self) -> str:
        url = self.upload_url + "/complete"
        try:
            req = Request(
                url,
                method="POST",
                headers={
                    **self.auth_headers,
                    "Accept": "application/json",
                    "Content-Type": "application/json",
                },
                data=json.dumps(
                    {
                        "parts": self._parts,
                    }
                ).encode(),
            )
            with _maybe_retry_request(req):
                pass
        except HTTPError as e:
            raise FileUploadException(
                f"Error completing upload {url}. Status {e.status}: {e.reason}"
            )

        return self.access_url

    @classmethod
    def save(
        cls,
        file: FileData,
        chunk_size: int | None = None,
        max_concurrency: int | None = None,
    ):
        import concurrent.futures

        multipart = cls(
            file.file_name,
            chunk_size=chunk_size,
            content_type=file.content_type,
            max_concurrency=max_concurrency,
        )
        multipart.create()

        parts = math.ceil(len(file.data) / multipart.chunk_size)
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=multipart.max_concurrency
        ) as executor:
            futures = []
            for part_number in range(1, parts + 1):
                start = (part_number - 1) * multipart.chunk_size
                data = file.data[start : start + multipart.chunk_size]
                futures.append(
                    executor.submit(multipart.upload_part, part_number, data)
                )

            for future in concurrent.futures.as_completed(futures):
                future.result()

        return multipart.complete()

    @classmethod
    def save_file(
        cls,
        file_path: str | Path,
        chunk_size: int | None = None,
        content_type: str | None = None,
        max_concurrency: int | None = None,
    ) -> str:
        import concurrent.futures

        file_name = os.path.basename(file_path)
        size = os.path.getsize(file_path)

        multipart = cls(
            file_name,
            chunk_size=chunk_size,
            content_type=content_type,
            max_concurrency=max_concurrency,
        )
        multipart.create()

        parts = math.ceil(size / multipart.chunk_size)
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=multipart.max_concurrency
        ) as executor:
            futures = []
            for part_number in range(1, parts + 1):

                def _upload_part(pn: int) -> None:
                    with open(file_path, "rb") as f:
                        start = (pn - 1) * multipart.chunk_size
                        f.seek(start)
                        data = f.read(multipart.chunk_size)
                        multipart.upload_part(pn, data)

                futures.append(executor.submit(_upload_part, part_number))

            for future in concurrent.futures.as_completed(futures):
                future.result()

        return multipart.complete()


@dataclass
class FalFileRepository(FalFileRepositoryBase):
    def save(
        self,
        file: FileData,
        multipart: bool | None = None,
        multipart_threshold: int | None = None,
        multipart_chunk_size: int | None = None,
        multipart_max_concurrency: int | None = None,
        object_lifecycle_preference: dict[str, str] | None = None,
    ) -> str:
        if multipart is None:
            threshold = multipart_threshold or MultipartUploadGCS.MULTIPART_THRESHOLD
            multipart = len(file.data) > threshold

        if multipart:
            return MultipartUploadGCS.save(
                file,
                chunk_size=multipart_chunk_size,
                max_concurrency=multipart_max_concurrency,
            )

        headers: Dict[str, str] = {}
        _object_lifecycle_headers(headers, object_lifecycle_preference)

        return self._save(file, "gcs", headers=headers)

    @property
    def auth_headers(self) -> dict[str, str]:
        token = fal_v3_token_manager.get_token()
        return {
            "Authorization": f"{token.token_type} {token.token}",
            "User-Agent": "fal/0.1.0",
        }

    def save_file(
        self,
        file_path: str | Path,
        content_type: str,
        multipart: bool | None = None,
        multipart_threshold: int | None = None,
        multipart_chunk_size: int | None = None,
        multipart_max_concurrency: int | None = None,
        object_lifecycle_preference: dict[str, str] | None = None,
    ) -> tuple[str, FileData | None]:
        if multipart is None:
            threshold = multipart_threshold or MultipartUploadGCS.MULTIPART_THRESHOLD
            multipart = os.path.getsize(file_path) > threshold

        if multipart:
            url = MultipartUploadGCS.save_file(
                file_path,
                chunk_size=multipart_chunk_size,
                content_type=content_type,
                max_concurrency=multipart_max_concurrency,
            )
            data = None
        else:
            with open(file_path, "rb") as f:
                data = FileData(
                    f.read(),
                    content_type=content_type,
                    file_name=os.path.basename(file_path),
                )
            url = self.save(
                data,
                object_lifecycle_preference=object_lifecycle_preference,
            )

        return url, data


class MultipartUpload:
    MULTIPART_THRESHOLD = 100 * 1024 * 1024
    MULTIPART_CHUNK_SIZE = 100 * 1024 * 1024
    MULTIPART_MAX_CONCURRENCY = 10

    def __init__(
        self,
        file_name: str,
        chunk_size: int | None = None,
        content_type: str | None = None,
        max_concurrency: int | None = None,
    ) -> None:
        self.file_name = file_name
        self.chunk_size = chunk_size or self.MULTIPART_CHUNK_SIZE
        self.content_type = content_type or "application/octet-stream"
        self.max_concurrency = max_concurrency or self.MULTIPART_MAX_CONCURRENCY

        self._parts: list[dict] = []

    def create(self):
        token = fal_v2_token_manager.get_token()
        try:
            req = Request(
                f"{token.base_upload_url}/upload/initiate-multipart",
                method="POST",
                headers={
                    "Authorization": f"{token.token_type} {token.token}",
                    "Accept": "application/json",
                    "Content-Type": "application/json",
                },
                data=json.dumps(
                    {
                        "file_name": self.file_name,
                        "content_type": self.content_type,
                    }
                ).encode(),
            )
            with _maybe_retry_request(req) as response:
                result = json.load(response)
                self._upload_url = result["upload_url"]
                self._file_url = result["file_url"]
        except HTTPError as exc:
            raise FileUploadException(
                f"Error initiating upload. Status {exc.status}: {exc.reason}"
            )

    def upload_part(self, part_number: int, data: bytes) -> None:
        url = f"{self._upload_url}&part_number={part_number}"

        req = Request(
            url,
            method="PUT",
            headers={"Content-Type": self.content_type},
            data=data,
        )

        try:
            with _maybe_retry_request(req, timeout=PUT_REQUEST_TIMEOUT) as resp:
                self._parts.append(
                    {
                        "part_number": part_number,
                        "etag": resp.headers["ETag"],
                    }
                )
        except HTTPError as exc:
            raise FileUploadException(
                f"Error uploading part {part_number} to {url}. "
                f"Status {exc.status}: {exc.reason}"
            )

    def complete(self):
        url = self._upload_url
        try:
            req = Request(
                url,
                method="POST",
                headers={
                    "Accept": "application/json",
                    "Content-Type": "application/json",
                },
                data=json.dumps({"parts": self._parts}).encode(),
            )
            with _maybe_retry_request(req):
                pass
        except HTTPError as e:
            raise FileUploadException(
                f"Error completing upload {url}. Status {e.status}: {e.reason}"
            )

        return self._file_url

    @classmethod
    def save(
        cls,
        file: FileData,
        chunk_size: int | None = None,
        max_concurrency: int | None = None,
    ):
        import concurrent.futures

        multipart = cls(
            file.file_name,
            chunk_size=chunk_size,
            content_type=file.content_type,
            max_concurrency=max_concurrency,
        )
        multipart.create()

        parts = math.ceil(len(file.data) / multipart.chunk_size)
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=multipart.max_concurrency
        ) as executor:
            futures = []
            for part_number in range(1, parts + 1):
                start = (part_number - 1) * multipart.chunk_size
                data = file.data[start : start + multipart.chunk_size]
                futures.append(
                    executor.submit(multipart.upload_part, part_number, data)
                )

            for future in concurrent.futures.as_completed(futures):
                future.result()

        return multipart.complete()

    @classmethod
    def save_file(
        cls,
        file_path: str | Path,
        chunk_size: int | None = None,
        content_type: str | None = None,
        max_concurrency: int | None = None,
    ) -> str:
        import concurrent.futures

        file_name = os.path.basename(file_path)
        size = os.path.getsize(file_path)

        multipart = cls(
            file_name,
            chunk_size=chunk_size,
            content_type=content_type,
            max_concurrency=max_concurrency,
        )
        multipart.create()

        parts = math.ceil(size / multipart.chunk_size)
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=multipart.max_concurrency
        ) as executor:
            futures = []
            for part_number in range(1, parts + 1):

                def _upload_part(pn: int) -> None:
                    with open(file_path, "rb") as f:
                        start = (pn - 1) * multipart.chunk_size
                        f.seek(start)
                        data = f.read(multipart.chunk_size)
                        multipart.upload_part(pn, data)

                futures.append(executor.submit(_upload_part, part_number))

            for future in concurrent.futures.as_completed(futures):
                future.result()

        return multipart.complete()


class MultipartUploadV3:
    MULTIPART_THRESHOLD = 100 * 1024 * 1024
    MULTIPART_CHUNK_SIZE = 10 * 1024 * 1024
    MULTIPART_MAX_CONCURRENCY = 10

    def __init__(
        self,
        file_name: str,
        chunk_size: int | None = None,
        content_type: str | None = None,
        max_concurrency: int | None = None,
    ) -> None:
        self.file_name = file_name
        self.chunk_size = chunk_size or self.MULTIPART_CHUNK_SIZE
        self.content_type = content_type or "application/octet-stream"
        self.max_concurrency = max_concurrency or self.MULTIPART_MAX_CONCURRENCY

        self._access_url: str | None = None
        self._upload_url: str | None = None

        self._parts: list[dict] = []

    @property
    def access_url(self) -> str:
        if not self._access_url:
            raise FileUploadException("Upload not initiated")
        return self._access_url

    @property
    def upload_url(self) -> str:
        if not self._upload_url:
            raise FileUploadException("Upload not initiated")
        return self._upload_url

    @property
    def auth_headers(self) -> dict[str, str]:
        fal_key = key_credentials()
        if not fal_key:
            raise FileUploadException("FAL_KEY must be set")

        key_id, key_secret = fal_key
        return {
            "Authorization": f"Key {key_id}:{key_secret}",
        }

    def create(self):
        grpc_host = os.environ.get("FAL_HOST", "api.alpha.fal.ai")
        rest_host = grpc_host.replace("api", "rest", 1)
        url = f"https://{rest_host}/storage/upload/initiate-multipart?storage_type=fal-cdn-v3"

        try:
            req = Request(
                url,
                method="POST",
                headers={
                    **self.auth_headers,
                    "Content-Type": "application/json",
                    "Accept": "application/json",
                },
                data=json.dumps(
                    {
                        "file_name": self.file_name,
                        "content_type": self.content_type,
                    }
                ).encode(),
            )

            with _maybe_retry_request(req) as response:
                result = json.load(response)
                self._access_url = result["file_url"]
                self._upload_url = result["upload_url"]

        except HTTPError as exc:
            raise FileUploadException(
                f"Error initiating upload. Status {exc.status}: {exc.reason}"
            )

    def upload_part(self, part_number: int, data: bytes) -> None:
        parsed = urlparse(self.upload_url)
        part_path = parsed.path + f"/{part_number}"
        url = urlunparse(parsed._replace(path=part_path))

        req = Request(
            url,
            method="PUT",
            headers={
                "Content-Type": self.content_type,
            },
            data=data,
        )

        try:
            with _maybe_retry_request(req, timeout=PUT_REQUEST_TIMEOUT) as resp:
                self._parts.append(
                    {
                        "partNumber": part_number,
                        "etag": resp.headers["ETag"],
                    }
                )
        except HTTPError as exc:
            raise FileUploadException(
                f"Error uploading part {part_number} to {url}. "
                f"Status {exc.status}: {exc.reason}"
            )

    def complete(self) -> str:
        parsed = urlparse(self.upload_url)
        complete_path = parsed.path + "/complete"
        url = urlunparse(parsed._replace(path=complete_path))

        try:
            req = Request(
                url,
                method="POST",
                headers={
                    "Accept": "application/json",
                    "Content-Type": "application/json",
                },
                data=json.dumps({"parts": self._parts}).encode(),
            )
            with _maybe_retry_request(req):
                pass
        except HTTPError as e:
            raise FileUploadException(
                f"Error completing upload {url}. Status {e.status}: {e.reason}"
            )

        return self.access_url

    @classmethod
    def save(
        cls,
        file: FileData,
        chunk_size: int | None = None,
        max_concurrency: int | None = None,
    ):
        import concurrent.futures

        multipart = cls(
            file.file_name,
            chunk_size=chunk_size,
            content_type=file.content_type,
            max_concurrency=max_concurrency,
        )
        multipart.create()

        parts = math.ceil(len(file.data) / multipart.chunk_size)
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=multipart.max_concurrency
        ) as executor:
            futures = []
            for part_number in range(1, parts + 1):
                start = (part_number - 1) * multipart.chunk_size
                data = file.data[start : start + multipart.chunk_size]
                futures.append(
                    executor.submit(multipart.upload_part, part_number, data)
                )

            for future in concurrent.futures.as_completed(futures):
                future.result()

        return multipart.complete()

    @classmethod
    def save_file(
        cls,
        file_path: str | Path,
        chunk_size: int | None = None,
        content_type: str | None = None,
        max_concurrency: int | None = None,
    ) -> str:
        import concurrent.futures

        file_name = os.path.basename(file_path)
        size = os.path.getsize(file_path)

        multipart = cls(
            file_name,
            chunk_size=chunk_size,
            content_type=content_type,
            max_concurrency=max_concurrency,
        )
        multipart.create()

        parts = math.ceil(size / multipart.chunk_size)
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=multipart.max_concurrency
        ) as executor:
            futures = []
            for part_number in range(1, parts + 1):

                def _upload_part(pn: int) -> None:
                    with open(file_path, "rb") as f:
                        start = (pn - 1) * multipart.chunk_size
                        f.seek(start)
                        data = f.read(multipart.chunk_size)
                        multipart.upload_part(pn, data)

                futures.append(executor.submit(_upload_part, part_number))

            for future in concurrent.futures.as_completed(futures):
                future.result()

        return multipart.complete()


class InternalMultipartUploadV3:
    MULTIPART_THRESHOLD = 100 * 1024 * 1024
    MULTIPART_CHUNK_SIZE = 10 * 1024 * 1024
    MULTIPART_MAX_CONCURRENCY = 10

    def __init__(
        self,
        file_name: str,
        chunk_size: int | None = None,
        content_type: str | None = None,
        max_concurrency: int | None = None,
    ) -> None:
        self.file_name = file_name
        self.chunk_size = chunk_size or self.MULTIPART_CHUNK_SIZE
        self.content_type = content_type or "application/octet-stream"
        self.max_concurrency = max_concurrency or self.MULTIPART_MAX_CONCURRENCY
        self._access_url: str | None = None
        self._upload_id: str | None = None

        self._parts: list[dict] = []

    @property
    def access_url(self) -> str:
        if not self._access_url:
            raise FileUploadException("Upload not initiated")
        return self._access_url

    @property
    def upload_id(self) -> str:
        if not self._upload_id:
            raise FileUploadException("Upload not initiated")
        return self._upload_id

    @property
    def auth_headers(self) -> dict[str, str]:
        token = fal_v3_token_manager.get_token()
        return {
            "Authorization": f"{token.token_type} {token.token}",
            "User-Agent": "fal/0.1.0",
        }

    def create(self):
        token = fal_v3_token_manager.get_token()
        try:
            req = Request(
                f"{token.base_upload_url}/files/upload/multipart",
                method="POST",
                headers={
                    **self.auth_headers,
                    "Accept": "application/json",
                    "Content-Type": self.content_type,
                    "X-Fal-File-Name": self.file_name,
                },
            )
            with _maybe_retry_request(req) as response:
                result = json.load(response)
                self._access_url = result["access_url"]
                self._upload_id = result["uploadId"]

        except HTTPError as exc:
            raise FileUploadException(
                f"Error initiating upload. Status {exc.status}: {exc.reason}"
            )

    def upload_part(self, part_number: int, data: bytes) -> None:
        url = f"{self.access_url}/multipart/{self.upload_id}/{part_number}"

        req = Request(
            url,
            method="PUT",
            headers={
                **self.auth_headers,
                "Content-Type": self.content_type,
            },
            data=data,
        )

        try:
            with _maybe_retry_request(req, timeout=PUT_REQUEST_TIMEOUT) as resp:
                self._parts.append(
                    {
                        "partNumber": part_number,
                        "etag": resp.headers["ETag"],
                    }
                )
        except HTTPError as exc:
            raise FileUploadException(
                f"Error uploading part {part_number} to {url}. "
                f"Status {exc.status}: {exc.reason}"
            )

    def complete(self) -> str:
        url = f"{self.access_url}/multipart/{self.upload_id}/complete"
        try:
            req = Request(
                url,
                method="POST",
                headers={
                    **self.auth_headers,
                    "Accept": "application/json",
                    "Content-Type": "application/json",
                },
                data=json.dumps({"parts": self._parts}).encode(),
            )
            with _maybe_retry_request(req):
                pass
        except HTTPError as e:
            raise FileUploadException(
                f"Error completing upload {url}. Status {e.status}: {e.reason}"
            )

        return self.access_url

    @classmethod
    def save(
        cls,
        file: FileData,
        chunk_size: int | None = None,
        max_concurrency: int | None = None,
    ):
        import concurrent.futures

        multipart = cls(
            file.file_name,
            chunk_size=chunk_size,
            content_type=file.content_type,
            max_concurrency=max_concurrency,
        )
        multipart.create()

        parts = math.ceil(len(file.data) / multipart.chunk_size)
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=multipart.max_concurrency
        ) as executor:
            futures = []
            for part_number in range(1, parts + 1):
                start = (part_number - 1) * multipart.chunk_size
                data = file.data[start : start + multipart.chunk_size]
                futures.append(
                    executor.submit(multipart.upload_part, part_number, data)
                )

            for future in concurrent.futures.as_completed(futures):
                future.result()

        return multipart.complete()

    @classmethod
    def save_file(
        cls,
        file_path: str | Path,
        chunk_size: int | None = None,
        content_type: str | None = None,
        max_concurrency: int | None = None,
    ) -> str:
        import concurrent.futures

        file_name = os.path.basename(file_path)
        size = os.path.getsize(file_path)

        multipart = cls(
            file_name,
            chunk_size=chunk_size,
            content_type=content_type,
            max_concurrency=max_concurrency,
        )
        multipart.create()

        parts = math.ceil(size / multipart.chunk_size)
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=multipart.max_concurrency
        ) as executor:
            futures = []
            for part_number in range(1, parts + 1):

                def _upload_part(pn: int) -> None:
                    with open(file_path, "rb") as f:
                        start = (pn - 1) * multipart.chunk_size
                        f.seek(start)
                        data = f.read(multipart.chunk_size)
                        multipart.upload_part(pn, data)

                futures.append(executor.submit(_upload_part, part_number))

            for future in concurrent.futures.as_completed(futures):
                future.result()

        return multipart.complete()


@dataclass
class FalFileRepositoryV2(FalFileRepositoryBase):
    def save(
        self,
        file: FileData,
        multipart: bool | None = None,
        multipart_threshold: int | None = None,
        multipart_chunk_size: int | None = None,
        multipart_max_concurrency: int | None = None,
        object_lifecycle_preference: dict[str, str] | None = None,
    ) -> str:
        if multipart is None:
            threshold = multipart_threshold or MultipartUpload.MULTIPART_THRESHOLD
            multipart = len(file.data) > threshold

        if multipart:
            return MultipartUpload.save(
                file,
                chunk_size=multipart_chunk_size,
                max_concurrency=multipart_max_concurrency,
            )

        token = fal_v2_token_manager.get_token()
        headers = {
            "Authorization": f"{token.token_type} {token.token}",
            "Accept": "application/json",
            "X-Fal-File-Name": file.file_name,
            "Content-Type": file.content_type,
        }

        _object_lifecycle_headers(headers, object_lifecycle_preference)

        storage_url = f"{token.base_upload_url}/upload"

        try:
            req = Request(
                storage_url,
                data=file.data,
                headers=headers,
                method="PUT",
            )
            with _maybe_retry_request(req, timeout=PUT_REQUEST_TIMEOUT) as response:
                result = json.load(response)

            return result["file_url"]
        except HTTPError as e:
            raise FileUploadException(
                f"Error initiating upload. Status {e.status}: {e.reason}"
            )

    def save_file(
        self,
        file_path: str | Path,
        content_type: str,
        multipart: bool | None = None,
        multipart_threshold: int | None = None,
        multipart_chunk_size: int | None = None,
        multipart_max_concurrency: int | None = None,
        object_lifecycle_preference: dict[str, str] | None = None,
    ) -> tuple[str, FileData | None]:
        if multipart is None:
            threshold = multipart_threshold or MultipartUpload.MULTIPART_THRESHOLD
            multipart = os.path.getsize(file_path) > threshold

        if multipart:
            url = MultipartUpload.save_file(
                file_path,
                chunk_size=multipart_chunk_size,
                content_type=content_type,
                max_concurrency=multipart_max_concurrency,
            )
            data = None
        else:
            with open(file_path, "rb") as f:
                data = FileData(
                    f.read(),
                    content_type=content_type,
                    file_name=os.path.basename(file_path),
                )
            url = self.save(
                data,
                object_lifecycle_preference=object_lifecycle_preference,
            )

        return url, data


@dataclass
class InMemoryRepository(FileRepository):
    def save(
        self,
        file: FileData,
        multipart: bool | None = None,
        multipart_threshold: int | None = None,
        multipart_chunk_size: int | None = None,
        multipart_max_concurrency: int | None = None,
        object_lifecycle_preference: dict[str, str] | None = None,
    ) -> str:
        return f'data:{file.content_type};base64,{b64encode(file.data).decode("utf-8")}'


@dataclass
class FalCDNFileRepository(FileRepository):
    def save(
        self,
        file: FileData,
        multipart: bool | None = None,
        multipart_threshold: int | None = None,
        multipart_chunk_size: int | None = None,
        multipart_max_concurrency: int | None = None,
        object_lifecycle_preference: dict[str, str] | None = None,
    ) -> str:
        headers = {
            **self.auth_headers,
            "Accept": "application/json",
            "Content-Type": file.content_type,
            "X-Fal-File-Name": file.file_name,
        }

        _object_lifecycle_headers(headers, object_lifecycle_preference)

        url = os.getenv("FAL_CDN_HOST", _FAL_CDN) + "/files/upload"
        request = Request(url, headers=headers, method="POST", data=file.data)
        try:
            with _maybe_retry_request(request) as response:
                result = json.load(response)
        except HTTPError as e:
            raise FileUploadException(
                f"Error initiating upload. Status {e.status}: {e.reason}"
            )

        access_url = result["access_url"]
        return access_url

    @property
    def auth_headers(self) -> dict[str, str]:
        key_creds = key_credentials()
        if not key_creds:
            raise FileUploadException("FAL_KEY must be set")

        key_id, key_secret = key_creds
        return {
            "Authorization": f"Bearer {key_id}:{key_secret}",
            "User-Agent": "fal/0.1.0",
        }


@dataclass
class FalFileRepositoryV3(FileRepository):
    @property
    def auth_headers(self) -> dict[str, str]:
        fal_key = key_credentials()
        if not fal_key:
            raise FileUploadException("FAL_KEY must be set")

        key_id, key_secret = fal_key
        return {
            "Authorization": f"Key {key_id}:{key_secret}",
            "User-Agent": "fal/0.1.0",
        }

    def save(
        self,
        file: FileData,
        multipart: bool | None = None,
        multipart_threshold: int | None = None,
        multipart_chunk_size: int | None = None,
        multipart_max_concurrency: int | None = None,
        object_lifecycle_preference: dict[str, str] | None = None,
    ) -> str:
        if multipart is None:
            threshold = multipart_threshold or MultipartUploadV3.MULTIPART_THRESHOLD
            multipart = len(file.data) > threshold

        if multipart:
            return MultipartUploadV3.save(
                file,
                chunk_size=multipart_chunk_size,
                max_concurrency=multipart_max_concurrency,
            )

        headers = {
            **self.auth_headers,
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
        _object_lifecycle_headers(headers, object_lifecycle_preference)

        grpc_host = os.environ.get("FAL_HOST", "api.alpha.fal.ai")
        rest_host = grpc_host.replace("api", "rest", 1)
        url = f"https://{rest_host}/storage/upload/initiate?storage_type=fal-cdn-v3"

        request = Request(
            url,
            headers=headers,
            method="POST",
            data=json.dumps(
                {
                    "file_name": file.file_name,
                    "content_type": file.content_type,
                }
            ).encode(),
        )
        try:
            with _maybe_retry_request(request) as response:
                result = json.load(response)
                file_url = result["file_url"]
                upload_url = result["upload_url"]
        except HTTPError as e:
            raise FileUploadException(
                f"Error initiating upload. Status {e.status}: {e.reason}"
            )

        request = Request(
            upload_url,
            headers={"Content-Type": file.content_type},
            method="PUT",
            data=file.data,
        )
        try:
            with _maybe_retry_request(request, timeout=PUT_REQUEST_TIMEOUT):
                pass
        except HTTPError as e:
            raise FileUploadException(
                f"Error uploading file. Status {e.status}: {e.reason}"
            )

        return file_url

    def save_file(
        self,
        file_path: str | Path,
        content_type: str,
        multipart: bool | None = None,
        multipart_threshold: int | None = None,
        multipart_chunk_size: int | None = None,
        multipart_max_concurrency: int | None = None,
        object_lifecycle_preference: dict[str, str] | None = None,
    ) -> tuple[str, FileData | None]:
        if multipart is None:
            threshold = multipart_threshold or MultipartUploadV3.MULTIPART_THRESHOLD
            multipart = os.path.getsize(file_path) > threshold

        if multipart:
            url = MultipartUploadV3.save_file(
                file_path,
                chunk_size=multipart_chunk_size,
                content_type=content_type,
                max_concurrency=multipart_max_concurrency,
            )
            data = None
        else:
            with open(file_path, "rb") as f:
                data = FileData(
                    f.read(),
                    content_type=content_type,
                    file_name=os.path.basename(file_path),
                )
            url = self.save(
                data,
                object_lifecycle_preference=object_lifecycle_preference,
            )

        return url, data


# This is only available for internal users to have long-lived access tokens
@dataclass
class InternalFalFileRepositoryV3(FileRepository):
    """
    InternalFalFileRepositoryV3 is a file repository that uses the FAL CDN V3.
    But generates and uses long-lived access tokens.
    That way it can avoid the need to refresh the token for every upload.
    """

    def save(
        self,
        file: FileData,
        multipart: bool | None = None,
        multipart_threshold: int | None = None,
        multipart_chunk_size: int | None = None,
        multipart_max_concurrency: int | None = None,
        object_lifecycle_preference: dict[str, str] | None = None,
    ) -> str:
        if multipart is None:
            threshold = (
                multipart_threshold or InternalMultipartUploadV3.MULTIPART_THRESHOLD
            )
            multipart = len(file.data) > threshold

        if multipart:
            return InternalMultipartUploadV3.save(
                file,
                chunk_size=multipart_chunk_size,
                max_concurrency=multipart_max_concurrency,
            )

        headers = {
            **self.auth_headers,
            "Accept": "application/json",
            "Content-Type": file.content_type,
            "X-Fal-File-Name": file.file_name,
        }

        _object_lifecycle_headers(headers, object_lifecycle_preference)

        url = os.getenv("FAL_CDN_V3_HOST", _FAL_CDN_V3) + "/files/upload"
        request = Request(url, headers=headers, method="POST", data=file.data)
        try:
            with _maybe_retry_request(request) as response:
                result = json.load(response)
        except HTTPError as e:
            raise FileUploadException(
                f"Error initiating upload. Status {e.status}: {e.reason}"
            )

        access_url = result["access_url"]
        return access_url

    @property
    def auth_headers(self) -> dict[str, str]:
        token = fal_v3_token_manager.get_token()
        return {
            "Authorization": f"{token.token_type} {token.token}",
            "User-Agent": "fal/0.1.0",
        }

    def save_file(
        self,
        file_path: str | Path,
        content_type: str,
        multipart: bool | None = None,
        multipart_threshold: int | None = None,
        multipart_chunk_size: int | None = None,
        multipart_max_concurrency: int | None = None,
        object_lifecycle_preference: dict[str, str] | None = None,
    ) -> tuple[str, FileData | None]:
        if multipart is None:
            threshold = (
                multipart_threshold or InternalMultipartUploadV3.MULTIPART_THRESHOLD
            )
            multipart = os.path.getsize(file_path) > threshold

        if multipart:
            url = InternalMultipartUploadV3.save_file(
                file_path,
                chunk_size=multipart_chunk_size,
                content_type=content_type,
                max_concurrency=multipart_max_concurrency,
            )
            data = None
        else:
            with open(file_path, "rb") as f:
                data = FileData(
                    f.read(),
                    content_type=content_type,
                    file_name=os.path.basename(file_path),
                )
            url = self.save(
                data,
                object_lifecycle_preference=object_lifecycle_preference,
            )

        return url, data



from __future__ import annotations

import shutil
import traceback
from functools import wraps
from pathlib import Path
from tempfile import NamedTemporaryFile, mkdtemp
from typing import Any, Callable, Optional
from urllib.parse import urlparse
from zipfile import ZipFile

import pydantic
from fastapi import Request

# https://github.com/pydantic/pydantic/pull/2573
if not hasattr(pydantic, "__version__") or pydantic.__version__.startswith("1."):
    IS_PYDANTIC_V2 = False
else:
    from pydantic import model_validator

    IS_PYDANTIC_V2 = True

from pydantic import BaseModel, Field

from fal.toolkit.file.providers.fal import (
    LIFECYCLE_PREFERENCE,
    FalCDNFileRepository,
    FalFileRepository,
    FalFileRepositoryV2,
    FalFileRepositoryV3,
    InMemoryRepository,
)
from fal.toolkit.file.providers.gcp import GoogleStorageRepository
from fal.toolkit.file.providers.r2 import R2Repository
from fal.toolkit.file.types import FileData, FileRepository, RepositoryId
from fal.toolkit.utils.download_utils import download_file

FileRepositoryFactory = Callable[[], FileRepository]

BUILT_IN_REPOSITORIES: dict[RepositoryId, FileRepositoryFactory] = {
    "fal": lambda: FalFileRepository(),
    "fal_v2": lambda: FalFileRepositoryV2(),
    "fal_v3": lambda: FalFileRepositoryV3(),
    "in_memory": lambda: InMemoryRepository(),
    "gcp_storage": lambda: GoogleStorageRepository(),
    "r2": lambda: R2Repository(),
    "cdn": lambda: FalCDNFileRepository(),
}


def get_builtin_repository(id: RepositoryId | FileRepository) -> FileRepository:
    if isinstance(id, FileRepository):
        return id

    if id not in BUILT_IN_REPOSITORIES.keys():
        raise ValueError(f'"{id}" is not a valid built-in file repository')
    return BUILT_IN_REPOSITORIES[id]()


get_builtin_repository.__module__ = "__main__"

DEFAULT_REPOSITORY: FileRepository | RepositoryId = "fal_v3"
FALLBACK_REPOSITORY: list[FileRepository | RepositoryId] = ["cdn", "fal"]
OBJECT_LIFECYCLE_PREFERENCE_KEY = "x-fal-object-lifecycle-preference"


@wraps(Field)
def FileField(*args, **kwargs):
    ui = kwargs.get("ui", {})
    ui.setdefault("field", "file")
    kwargs["ui"] = ui
    return Field(*args, **kwargs)


def _try_with_fallback(
    func: str,
    args: list[Any],
    repository: FileRepository | RepositoryId,
    fallback_repository: Optional[
        FileRepository | RepositoryId | list[FileRepository | RepositoryId]
    ],
    save_kwargs: dict,
    fallback_save_kwargs: dict,
) -> Any:
    if fallback_repository is None:
        fallback_repository = []
    elif isinstance(fallback_repository, list):
        pass
    else:
        fallback_repository = [fallback_repository]

    attempts: list[tuple[FileRepository | RepositoryId, dict]] = [
        (repository, save_kwargs),
        *((fallback, fallback_save_kwargs) for fallback in fallback_repository),
    ]
    for idx, (repo, kwargs) in enumerate(attempts):
        repo_obj = get_builtin_repository(repo)
        try:
            return getattr(repo_obj, func)(*args, **kwargs)
        except Exception as exc:
            if idx >= len(attempts) - 1:
                raise

            traceback.print_exc()
            print(
                f"Failed to {func} to repository {repo}: {exc}, "
                f"falling back to {attempts[idx + 1][0]}"
            )


class File(BaseModel):
    # public properties
    url: str = Field(
        description="The URL where the file can be downloaded from.",
    )
    content_type: Optional[str] = Field(
        None,
        description="The mime type of the file.",
        examples=["image/png"],
    )
    file_name: Optional[str] = Field(
        None,
        description="The name of the file. It will be auto-generated if not provided.",
        examples=["z9RV14K95DvU.png"],
    )
    file_size: Optional[int] = Field(
        None, description="The size of the file in bytes.", examples=[4404019]
    )
    file_data: Optional[bytes] = Field(
        None,
        description="File data",
        exclude=True,
        repr=False,
    )

    # Pydantic custom validator for input type conversion
    if IS_PYDANTIC_V2:

        @model_validator(mode="before")
        @classmethod
        def __convert_from_str_v2(cls, value: Any):
            if isinstance(value, str):
                parsed_url = urlparse(value)
                if parsed_url.scheme not in ["http", "https", "data"]:
                    raise ValueError("value must be a valid URL")
                # Return a mapping so the model can be constructed normally
                return {"url": parsed_url.geturl()}
            return value

    else:

        @classmethod
        def __convert_from_dict(cls, value: Any):
            if isinstance(value, dict):
                return cls(**value)
            return value

        @classmethod
        def __get_validators__(cls):
            yield cls.__convert_from_dict
            yield cls.__convert_from_str

    @classmethod
    def __convert_from_str(cls, value: Any):
        if isinstance(value, str):
            parsed_url = urlparse(value)
            if parsed_url.scheme not in ["http", "https", "data"]:
                raise ValueError("value must be a valid URL")
            return cls._from_url(parsed_url.geturl())

        return value

    @classmethod
    def _from_url(
        cls,
        url: str,
    ) -> File:
        return cls(
            url=url,
            content_type=None,
            file_name=None,
            file_size=None,
            file_data=None,
        )

    @classmethod
    def from_bytes(
        cls,
        data: bytes,
        content_type: Optional[str] = None,
        file_name: Optional[str] = None,
        repository: FileRepository | RepositoryId = DEFAULT_REPOSITORY,
        fallback_repository: Optional[
            FileRepository | RepositoryId | list[FileRepository | RepositoryId]
        ] = FALLBACK_REPOSITORY,
        request: Optional[Request] = None,
        save_kwargs: Optional[dict] = None,
        fallback_save_kwargs: Optional[dict] = None,
    ) -> File:
        save_kwargs = save_kwargs or {}
        fallback_save_kwargs = fallback_save_kwargs or {}

        fdata = FileData(data, content_type, file_name)

        if request:
            object_lifecycle_preference = request_lifecycle_preference(request)
        else:
            print("[WARNING] No request provided, using global lifecycle preference")
            object_lifecycle_preference = LIFECYCLE_PREFERENCE.get()

        save_kwargs.setdefault(
            "object_lifecycle_preference", object_lifecycle_preference
        )
        fallback_save_kwargs.setdefault(
            "object_lifecycle_preference", object_lifecycle_preference
        )

        url = _try_with_fallback(
            "save",
            [fdata],
            repository=repository,
            fallback_repository=fallback_repository,
            save_kwargs=save_kwargs,
            fallback_save_kwargs=fallback_save_kwargs,
        )

        return cls(
            url=url,
            content_type=fdata.content_type,
            file_name=fdata.file_name,
            file_size=len(data),
            file_data=data,
        )

    @classmethod
    def from_path(
        cls,
        path: str | Path,
        content_type: Optional[str] = None,
        repository: FileRepository | RepositoryId = DEFAULT_REPOSITORY,
        multipart: bool | None = None,
        fallback_repository: Optional[
            FileRepository | RepositoryId | list[FileRepository | RepositoryId]
        ] = FALLBACK_REPOSITORY,
        request: Optional[Request] = None,
        save_kwargs: Optional[dict] = None,
        fallback_save_kwargs: Optional[dict] = None,
    ) -> File:
        file_path = Path(path)
        if not file_path.exists():
            raise FileNotFoundError(f"File {file_path} does not exist")

        save_kwargs = save_kwargs or {}
        fallback_save_kwargs = fallback_save_kwargs or {}

        content_type = content_type or "application/octet-stream"

        if request:
            object_lifecycle_preference = request_lifecycle_preference(request)
        else:
            print("[WARNING] No request provided, using global lifecycle preference")
            object_lifecycle_preference = LIFECYCLE_PREFERENCE.get()

        save_kwargs.setdefault(
            "object_lifecycle_preference", object_lifecycle_preference
        )
        fallback_save_kwargs.setdefault(
            "object_lifecycle_preference", object_lifecycle_preference
        )

        save_kwargs.setdefault("multipart", multipart)
        fallback_save_kwargs.setdefault("multipart", multipart)

        save_kwargs.setdefault("content_type", content_type)
        fallback_save_kwargs.setdefault("content_type", content_type)

        url, data = _try_with_fallback(
            "save_file",
            [file_path],
            repository=repository,
            fallback_repository=fallback_repository,
            save_kwargs=save_kwargs,
            fallback_save_kwargs=fallback_save_kwargs,
        )

        return cls(
            url=url,
            file_data=data.data if data else None,
            content_type=content_type,
            file_name=file_path.name,
            file_size=file_path.stat().st_size,
        )

    def as_bytes(self) -> bytes:
        if self.file_data is None:
            raise ValueError("File has not been downloaded")

        return self.file_data

    def save(self, path: str | Path, overwrite: bool = False) -> Path:
        file_path = Path(path).resolve()

        if file_path.exists() and not overwrite:
            raise FileExistsError(f"File {file_path} already exists")

        downloaded_path = download_file(self.url, target_dir=file_path.parent)
        downloaded_path.rename(file_path)

        return file_path


class CompressedFile(File):
    extract_dir: Optional[str] = Field(default=None, exclude=True, repr=False)

    def __iter__(self):
        if not self.extract_dir:
            self._extract_files()

        files = Path(self.extract_dir).iterdir()  # type: ignore
        return iter(files)

    def _extract_files(self):
        self.extract_dir = mkdtemp()

        with NamedTemporaryFile() as temp_file:
            file_path = temp_file.name
            self.save(file_path, overwrite=True)

            with ZipFile(file_path) as zip_file:
                zip_file.extractall(self.extract_dir)

    def glob(self, pattern: str):
        if not self.extract_dir:
            self._extract_files()

        return Path(self.extract_dir).glob(pattern)  # type: ignore

    def __del__(self):
        if self.extract_dir:
            shutil.rmtree(self.extract_dir)


def request_lifecycle_preference(request: Request) -> dict[str, str] | None:
    import json

    preference_str = request.headers.get(OBJECT_LIFECYCLE_PREFERENCE_KEY)
    if preference_str is None:
        return None

    try:
        return json.loads(preference_str)
    except Exception as e:
        print(f"Failed to parse object lifecycle preference: {e}")
        return None



from __future__ import annotations

from typing import Any

import structlog
from structlog.typing import EventDict, WrappedLogger

from .style import LEVEL_STYLES


class DebugConsoleLogProcessor:
    """
    Log processor that prints to the console the well-formatted
    messages only when in debug mode (enabled by the `--debug` flag).
    """

    debug: bool = False
    renderer = structlog.dev.ConsoleRenderer(level_styles=LEVEL_STYLES)

    def __call__(self, logger: WrappedLogger, method_name: str, event_dict: EventDict):
        if self.debug:
            return self.renderer.__call__(logger, method_name, event_dict)
        raise structlog.DropEvent


_console_log_output = DebugConsoleLogProcessor()


def set_debug_logging(debug: bool):
    """Enable/disable console log output."""
    _console_log_output.debug = debug


structlog.configure(
    processors=[
        structlog.stdlib.add_log_level,
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M:%S"),
        structlog.processors.StackInfoRenderer(),
        _console_log_output,
    ],
    wrapper_class=structlog.stdlib.BoundLogger,
)


def get_logger(*args: Any, **initial_values: Any) -> structlog.stdlib.BoundLogger:
    return structlog.get_logger(*args, **initial_values)



import argparse
import json

from fal.api.client import SyncServerlessClient

from .parser import FalClientParser, RefAction, get_output_parser


def _deploy(args):
    from ._utils import get_app_data_from_toml, is_app_name

    team = args.team
    app_ref = args.app_ref

    # If the app_ref is an app name, get team from pyproject.toml
    if app_ref and is_app_name(app_ref):
        try:
            *_, toml_team = get_app_data_from_toml(app_ref[0])
            team = team or toml_team
        except (ValueError, FileNotFoundError):
            # If we can't find the app in pyproject.toml, team remains None
            pass

    client = SyncServerlessClient(host=args.host, team=team)
    res = client.deploy(
        app_ref,
        app_name=args.app_name,
        auth=args.auth,
        strategy=args.strategy,
        reset_scale=args.app_scale_settings,
        force_env_build=args.force_env_build,
    )
    app_id = res.revision
    resolved_app_name = res.app_name

    if args.output == "json":
        args.console.print(
            json.dumps({"revision": app_id, "app_name": resolved_app_name})
        )
    elif args.output == "pretty":
        from fal.flags import URL_OUTPUT

        args.console.print(
            "Registered a new revision for function "
            f"'{resolved_app_name}' (revision='{app_id}')."
        )
        if URL_OUTPUT != "none":
            args.console.print("Playground:")
            for url in res.urls.get("playground", {}).values():
                args.console.print(f"\t{url}")
        if URL_OUTPUT == "all":
            args.console.print("Synchronous Endpoints:")
            for url in res.urls.get("sync", {}).values():
                args.console.print(f"\t{url}")
            args.console.print("Asynchronous Endpoints (Recommended):")
            for url in res.urls.get("async", {}).values():
                args.console.print(f"\t{url}")
    else:
        raise AssertionError(f"Invalid output format: {args.output}")


def add_parser(main_subparsers, parents):
    from fal.sdk import ALIAS_AUTH_MODES

    def valid_auth_option(option):
        if option not in ALIAS_AUTH_MODES:
            raise argparse.ArgumentTypeError(f"{option} is not a auth option")
        return option

    deploy_help = (
        "Deploy a fal application. "
        "If no app reference is provided, the command will look for a "
        "pyproject.toml file with a [tool.fal.apps] section and deploy the "
        "application specified with the provided app name."
    )

    epilog = (
        "Examples:\n"
        "  fal deploy\n"
        "  fal deploy path/to/myfile.py\n"
        "  fal deploy path/to/myfile.py::MyApp\n"
        "  fal deploy path/to/myfile.py::MyApp --app-name myapp --auth public\n"
        "  fal deploy my-app\n"
    )

    parser = main_subparsers.add_parser(
        "deploy",
        parents=[
            *parents,
            get_output_parser(),
            FalClientParser(add_help=False),
        ],
        description=deploy_help,
        help=deploy_help,
        epilog=epilog,
    )

    parser.add_argument(
        "app_ref",
        nargs="?",
        action=RefAction,
        help=(
            "Application reference. Either a file path or a file path and a "
            "function name separated by '::'. If no reference is provided, the "
            "command will look for a pyproject.toml file with a [tool.fal.apps] "
            "section and deploy the application specified with the provided app name.\n"
            "File path example: path/to/myfile.py::MyApp\n"
            "App name example: my-app (configure team in pyproject.toml)\n"
        ),
    )

    parser.add_argument(
        "--app-name",
        help="Application name to deploy with.",
    )

    parser.add_argument(
        "--auth",
        type=valid_auth_option,
        help="Application authentication mode (private, public).",
    )
    parser.add_argument(
        "--strategy",
        choices=["recreate", "rolling"],
        help="Deployment strategy.",
        default="rolling",
    )
    parser.add_argument(
        "--no-scale",
        action="store_false",
        dest="app_scale_settings",
        default=False,
        help="Use the previous deployment of the application for scale settings. "
        "This is the default behavior.",
    )
    parser.add_argument(
        "--reset-scale",
        action="store_true",
        dest="app_scale_settings",
        help="Use the application code for scale settings.",
    )
    parser.add_argument(
        "--force-env-build",
        action="store_true",
        help="Ignore the environment build cache and force rebuild.",
    )

    parser.set_defaults(func=_deploy)
